"use strict"

/*
 * Copyright (c) 2020, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

// Global SCB object - for storing information that cannot otherwise be implemented in the codec
let SCB = {
	codecInstances: [],
	bulkDeviceConnected: false,
	bulkDeviceKey: '',
	connectedHwInfo: {},

	// Shortcut that returns the first connected codec instance
	codec: function () {
		return this.codecInstances.filter(e => e._isCodecConnected)[0];
	},
	enableLogging: function () {
		if (this.codec()) { this.codec().logSerialData = true; }
		if (gc.desktop.isDesktop()) {
			gc.console.setCookie("scbCustomCodec.js", 4);
			gc.console.setCookie("ti-device-identifier", 5);
		}
	}
};

//Global variable
let EVMID_OLD = -1; //Global to used to track EVM ID

/*
 * gc global variable provides access to GUI Composer infrastructure components and project information.
 * For more information, please see the Working with Javascript guide in the online help.
 */
var gc = gc || {};
gc.services = gc.services || {};


/**
 * This code implements the IPacketCodec interface.
 *
 * @constructor
 * @implements gc.databind.IPacketCodec
 */
var CustomCodec = function () { };

var initCodec = function () {

	// Enabling logging
	const moduleName = "scbCustomCodec.js";
	if (gc.desktop.isDesktop()) { gc.console.setCookie(moduleName, 2); }
	gc.console.debug(moduleName, "Initializing codec...");

	// IMPORTANT: Do NOT add any objects to the CustomCodec's protoype before calling the
	// gc.databind.AbstractMessageBasedCodec(), as they will get overwritten!

	// Add commandQueue to this instance (Refer to GC-2502)
	CustomCodec = function () {
		gc.databind.AbstractMessageBasedCodec.call(this, "SCB");
	};

	//////////////////////// AbstractMessageBasedCodec APIs ////////////////////////

	// See additionals APIs and object provided by 'ti-core-databind\src\AbstractMessageBasedCodec.js'
	CustomCodec.prototype = new gc.databind.AbstractMessageBasedCodec("SCB");
	// Objects: name, commandQueue, disconnect()

	// Overwriting the default 'connect' method in AbstractMessageBasedCodec...
	CustomCodec.prototype.connect = function (controllerConfig) {

		gc.console.debug(moduleName, "Calling connect()...");

		// Update the function scope for the AbstractMessageBasedCodec connect method
		gc.databind.AbstractMessageBasedCodec.prototype.connect.call(this);

		// Get controllerConfig (framework provides this as array) - CURRENTLY UNUSED BY CODEC
		// controllerConfig = Array.isArray(controllerConfig) ? controllerConfig[0] : controllerConfig;

		// Initialize codec instance
		this.initialize();

		// Save codec instance
		SCB.codecInstances.push(this);

		// IMPORTANT: The connection process will NOT be completed until we've returned a resolved promise here!
		return Q();
	};

	// Overwriting the default 'disconnect' method in AbstractMessageBasedCodec...
	CustomCodec.prototype.disconnect = function () {

		gc.console.debug(moduleName, "Calling disconnect()...");

		// Update the function scope for the AbstractMessageBasedCodec disconnect method
		gc.databind.AbstractMessageBasedCodec.prototype.disconnect.call(this);

		// Set connection flag
		this.serialDeviceConnected = false;

		// Clear all pending commands and deferred writeValue and readValue promises
		//this.pendingSendCmds = [];
		//this.commandQueue.clearAll();   // This is handled by the AbstractMessageBasedCodec's disconnect method
		clearTimeout(this.connectionTimeOut);
		//this.strTxBuffer = "";
		//this.strRxBuffer = "";

		// Disconnect from bulk device
		if (SCB.bulkDeviceConnected) {
			closeBulkDevice(false); 	// 'false' indicates that this is not in response to a detach event
		}

		// Remove matching instance
		let index = SCB.codecInstances.indexOf(this);
		if (index >= 0) { SCB.codecInstances.splice(index, 1); }
	};

	// Overwriting the default 'isStillConnected' method in AbstractMessageBasedCodec...
	// This function is periodically called by the internal state machine to check if hardware is connected.
	CustomCodec.prototype.isStillConnected = function () {

		// (OPTIONAL) Ping the hardware and check for a response

		// Resolve promise based on our internal state flag
		if (this.serialDeviceConnected) { return Q.resolve(true); }
		else { return Q.reject(false); }
	};


	//////////////////////// Connection helper functions ////////////////////////

	// Requests ID information from device 
	CustomCodec.prototype.identify = async function () {

		let info = {};

		// NOTE: try/catch is used to skip awaiting all commands in case the device is not responding
		try {

			// Get firmware ID
			let jsonResponse = await this.sendCommand(EVM_ID_COMMAND, DEFAULT_CMDTIMEOUT_MS);
			info.firmwareInfo = this.parseFirmwareInfo(jsonResponse);

			// Get Controller ID (from MSP432E EEPROM)
			info.controllerInfo = await this.readEEPROM(true);

			// Get EVM ID (from EVM EEPROM)
			//info.evmInfo = await this.readEEPROM();
            //Have no EEPROM, so use values from FirmwareInfo: 
            info.evmInfo = JSON.parse(JSON.stringify(info.controllerInfo));
            
            info.evmInfo.brd_name = info.firmwareInfo.name;
            info.evmInfo.dev_name = info.firmwareInfo.name.replace("EVM", "");
            info.evmInfo.brd_rev = "N/A";
            info.evmInfo.brd_sno = "N/A";
            
            //initialze globals
            if(info.evmInfo.dev_name === "INA229" || info.evmInfo.dev_name === "INA239")
            {
                isSPI = true;
            }
            else
            {
                isSPI = false;
            }
            if(info.evmInfo.dev_name === "INA229" || info.evmInfo.dev_name === "INA228")
            {
                is20Bit = true;
            }
            else
            {
                is20Bit = false;
            }
            
            //Get default switch setting
            jsonResponse = await this.sendCommand("getdevice", DEFAULT_CMDTIMEOUT_MS);
            if(jsonResponse.ChipSelect)
            {
                EVMID_OLD = jsonResponse.ChipSelect;
                if(templateObj.$.ti_widget_droplist_E1_SW1.selectedValue === 0)
                {
                    templateObj.$.ti_widget_droplist_E1_SW1.selectedValue = jsonResponse.ChipSelect;
                }
            }
            else if(jsonResponse.Address)
            {
                EVMID_OLD = (jsonResponse.Address) & 0xF;
                if(templateObj.$.ti_widget_droplist_E1_SW1.selectedValue === 0 |templateObj.$.ti_widget_droplist_E1_SW0.selectedValue === 0)
                {
                    templateObj.$.ti_widget_droplist_E1_SW1.selectedValue = ((jsonResponse.Address >> 2) & 0x3) + 1;
                    templateObj.$.ti_widget_droplist_E1_SW0.selectedValue = (jsonResponse.Address & 0x3) + 1;
                }
            }
            else
            {
                //error...
            }

			// CONSIDER: Requesting a list of supported commands - Must figure out how to make this information sticky
			//jsonResponse = await this.sendCommand("help", DEFAULT_CMDTIMEOUT_MS);
			//jsonResponse = await this.sendCommand("help showHidden", DEFAULT_CMDTIMEOUT_MS);

		} catch (error) {
			console.error(moduleName, `Error in CustomCodec.identify(): ${error}`);
		}

		return info;
	};

	CustomCodec.prototype.verifyRegister = async function (verifyRegisterObj) {

		let jsonResponse, readValue;
		if (verifyRegisterObj) {
			const addressString = gc.utils.string2value(verifyRegisterObj.addr).toString(16);

			// CONSIDER: Changing this to a dedicated readRegister function at some point
			jsonResponse = await this.sendCommand(READ_REGISTER_COMMAND + " " + addressString, DEFAULT_CMDTIMEOUT_MS);
			if (!jsonResponse.register) {
				console.error(moduleName, `Received unexpected response: ${jsonResponse}`);
			} else {
				readValue = jsonResponse.register.value;
				verifyRegisterObj.registerValue = "0x" + readValue.toString(16);

				const value = +verifyRegisterObj.value || 0;
				const mask = verifyRegisterObj.mask === undefined ? -1 : +verifyRegisterObj.mask;
				verifyRegisterObj.result = ((readValue & mask) === value);
			}
			// if (!verifyRegisterObj.result) {
			// 	console.error(moduleName, `Incorrect value (${readValue}) read from register at address = 0x${addressString}`);
			// }
			return verifyRegisterObj;
		}
	};

	// Update firmware
	CustomCodec.prototype.updateFirmware = async function () {

		// Install DFU driver
		await Utilities.wrapQ(Utilities.installDfuDriver());

		// Send 'bsl' command
		let promise = this.addCommand(BSL_COMMAND, this.pendingSendCmds++);
		this.encoder(BSL_COMMAND);

		// Set a command timeout
		this.connectionTimeOut = setTimeout(() => {
			this.resolveCmdPromise(new RegExp(BSL_COMMAND), "Device did not respond to 'bsl' command", false);
			clearTimeout(this.connectionTimeOut);
		}, DEFAULT_CMDTIMEOUT_MS);

		// Wait for acknowledge or timeout to occur, then clear the timeout
		let ack = await Utilities.wrapQ(promise);
		// clearTimeout(this.connectionTimeOut);

		// If device acknowledged, call Utilities.bsl_update();
		if (ack) { Utilities.bsl_update(); }
		else { console.error(moduleName, "Device did not acknowledge 'bsl' command"); }
	}


	/////////////////////////// IPollingPacketCodec APIs ///////////////////////////

	// NOTE: AbstractMessageBasedCodec also implements the IPollingPacketCodec
	// CustomCodec.prototype = new gc.databind.IPollingPacketCodec();

	const jsonDelimiter = '\r\n';

	/**
	 * Encodes data into a packet for sending to the target device.
	 *
	 * @param target - function to call to pass on the encoded data towards the target
	 * @param jsonObj - The data object from the GUI that is to be encoded into a packet for sending.
	 */
	CustomCodec.prototype.encode = function (target, cmdString) {

		// Do not send commands while EVM is busy.
		if (this.currentState === "busy") {

			// Enque command
			this.enqueueCmd(target, cmdString);
			return;

// 		} else if ((this.currentState === "collecting") && (cmdString !== STOP_CONV_COMMAND)) {

// 			// Command not allowed while collecting data
// 			return;
		}

		// Send command...
		this.strTxBuffer = cmdString; // We will use this buffer for recalling the last command sent to the device
		if (this.strTxBuffer.length > 0) {

			// Update EVM state to busy
			this.setState("busy");

			// Log message (NOTE: Separator is used to make the log easier to read)
			if (this.logSerialData) {
				console.info(moduleName, "------------------------------------------");
				console.info(moduleName, `Sending command '${this.strTxBuffer}'`);
			}

			try {
				target(this.strTxBuffer);
			} catch (ex) {
				console.error(moduleName, "encode exception: " + ex);
			}
		}
	};


	/**
	 * Decodes packets from the target device into data objects.  One object for each packet of data.
	 * Unless you have chained your custom codec with the CR codec, the packet data is not framed,
	 * so this method is responsible for decoding partial packets as well as multiple packets.  Partial packets
	 * should not be returned, but deferred until the next decode method call with more raw data to process.  As a
	 * result, this method may return 0 or more packets of data on each invocation.
	 *
	 * @param target -  function to call to pass on a JSON object containing the decoded values towards the GUI
	 * @param data  - The raw data received that needs to be decoded from packets into objects.
	 * @return - true if connected (i.e. valid data received), false if not connected
	 */
	CustomCodec.prototype.decode = function (target, data) {

		// Convert data to ASCII
		var strRxData = data.map(value => String.fromCharCode(value)).join("");
		clearTimeout(this.connectionTimeOut);

		// Add RX data to strRxBuffer and look for a newline character...
		this.strRxBuffer = this.strRxBuffer + strRxData;
		while (this.strRxBuffer.includes(jsonDelimiter)) {

			// Parse JSON message from strRxBuffer (message must always end with 'jsonDelimiter' string!)
			let index = this.strRxBuffer.indexOf(jsonDelimiter);
			const strMessage = this.strRxBuffer.substring(0, index)	// Delimiter is NOT included in strMessage
			if (this.logSerialData) {
				console.info(moduleName, "Received response " + strMessage);
			}

			// Remove JSON message from strRxBuffer
			this.strRxBuffer = this.strRxBuffer.substring(strMessage.length + jsonDelimiter.length);	// Removes strMessage AND delimiter

			// Check for a message to parse
			if (!strMessage) { break; }

			// Try to parse message as a JSON object
			try {

				// Parse JSON string to object
				var jsonObj = JSON.parse(strMessage.trim());


				//
				// Look for predefined JSON properties to decide what action(s) to take...
				//

				// Save the parsed message as the last recieved object (expect when object is notifying us of a state change)
				// This allows us to return data from the sendCommand() function.
				if (!jsonObj.hasOwnProperty("evm_state")) { this.lastMessage = jsonObj; }

				// Loop through all properties in the JSON object...
				for (const property in jsonObj) {

					// Handle each case...
					switch (property) {

						// EVM state (tells us if EVM is 'idle' and ready to receive commands)
						case "evm_state":	// CONSIDER: Shortening this to "state"
							const newState = jsonObj.evm_state;		// {"evm_state":"idle"}
							this.setState(newState);
							break;


						// EVM firmware is acknowledging a received command...
						case "acknowledge":	// CONSIDER: Shortening this to "ack"
							const cmd_ack = jsonObj.acknowledge;		// {"acknowledge":"rreg 0"}

							// Check if this response does not match the last command sent to the EVM...
							if (cmd_ack !== this.strTxBuffer) {
								console.error(moduleName, `The EVM's acknowledged command ('${cmd_ack}') 
								does not match the last command sent ('${this.strTxBuffer}')`);
							}

							// If BSL command is acknowledged, resolve the 'bsl' command promise
							if (cmd_ack === BSL_COMMAND) {
								this.resolveCmdPromise(new RegExp(BSL_COMMAND), true);
							}
							
							// If setdevice command is acknowledged, resolve the 'setdevice' command promise
							if (cmd_ack.includes('setdevice')) {
								this.resolveCmdPromise(new RegExp('setdevice'), true);
							}
							break;


						// Register data
						case "register":
							const readValue = jsonObj.register.value;	// {"register":{"address":0,"value":129}}

							// Check if we are resolving a register model promise...
							let resolved = this.resolveCmdPromise(/^rreg/, +readValue);
							if (!resolved) {
								// Update the register model directly (to keep register model up-to-date)
								this.setRegisterValue(jsonObj.register.address, readValue);
							}
							break;


						// Log firmware message to console & eventlog...
						case "console":
							const msg = jsonObj.console;				// {"console":"This is a console message..."}
							console.log(moduleName, msg);
							if (this.eventLog) { this.eventLog.log("info", msg); }
							break;


						// Log error message to console & eventlog...
						case "error":
							const errorNo = jsonObj.error.no;			// {"error":{"no":3,"msg":"I2C EEPROM read failed"}}
							const errorMsg = jsonObj.error.msg;

							console.error(errorMsg, "\nErrorNo: ", errorNo)
							if (this.eventLog) { this.eventLog.log("error", errorMsg); }

							// Special error handlers
							// const main = document.querySelector('ti-sar-main');
							// if (errorNo === 5) { // ERROR_BAD_CRC_BYTE
							// 	main.showClearCRCErrorPopUp();
							// }

							break;

						// case "interrupt":
						// 	console.log("################-------> INTERRUPT");
						// 	this.interruptPacket = jsonObj;
						// 	if (this.dataCapture) {
						// 		this.handleCaptureInterrupt = true;
						// 	}
						// 	else {
						// 		this.handleInterrupt();
						// 	}

					}
				}

				// Send the received and decoded javascript object to the application where you can use it to bind to widget properties
				//target(jsonObj);

			} catch (ex) {
				console.error(moduleName, `decoded '${strMessage}' from target, ${ex}`);
			}
		}

		// Return true to set the 'Hardware Connected' status in the status bar
		return this.serialDeviceConnected;
	};


	/////////////////////////
	// Register model APIs //
	/////////////////////////

	/*
	* Confluence: https://confluence.itg.ti.com/pages/viewpage.action?pageId=173550282
	* For an example of a codec that accesses symbols using readValue/writeValue,
	* please see components/ti-widget-registers/demo/MotorwareDemo_F28069M/customCodec_0.js
	*/

	// Called by register model to read a device register...
	// IMPORTANT: This function is ONLY intended to be called by the register model.
	// Should you need to read a register, use CustomCodec.readRegister() instead!
	CustomCodec.prototype.readValue = function (regInfo) {
        
		// Get the address and format as a hex string (with no prefix or postfix)
		const { addr } = regInfo;

		// (OPTIONAL) Ignore register readback under certain conditions
		// const ignoreAddress = 0x07;
		// const ignoreMask = 0x08;
		// const ignoreValue = 0x08;
		// if (regInfo.addr === ignoreAddress) {
		// 	const currentRegValue = this.getRegisterValue(ignoreAddress);
		// 	if ((currentRegValue & ignoreMask) === ignoreValue) {
		// 		return currentRegValue;
		// 	}
		// }

		// Call read register function
		// NOTE: Wrap native JS promise in Q promise
		return Utilities.wrapJS(this.readRegister(addr, selectedEVM), "register.value");
	};

	// Called by register model to write register data...
	CustomCodec.prototype.writeValue = function (regInfo, value) {
 
		// Build the write register command string
		const { addr } = regInfo;

		// Send write register command to queue the command locally
		// NOTE 1: writeRegister() returns a native JS promise so we must wrap it in a Q promise.
		// NOTE 2: The register model does not look at the resolved promise value.
		return Utilities.wrapJS(this.writeRegister(addr, value, selectedEVM));
	};


	////////////////////// Register model helper functions //////////////////////

	// Reads register at given address. Address argument can be either a hex string or an integer.
	CustomCodec.prototype.readRegister = function (address, coreIndex) {
        
        //Set EVM ID
        if(typeof this.sendRegmapEVMID === "function")
        {
            this.sendRegmapEVMID(coreIndex);
        }

		// Get integer address
		const intAddr = (typeof address === 'string') ? parseInt(address, 16) : address;
		if (isNaN(intAddr)) {
			console.error(moduleName, `Invalid register address provided to readRegister(): ${address}`)
			return;
		}

		// Build and send read register command string
		const cmd = READ_REGISTER_COMMAND + ' ' + intAddr.toString(16);
		let jsPromise = this.sendCommand(cmd, DEFAULT_CMDTIMEOUT_MS);

		return jsPromise;	// Resolves to an object, such as: {"register":{"address":0,"value":129}}
	};

	// Writes to a register, given an address and value. Arguments can be either a hex string or an integer.
	CustomCodec.prototype.writeRegister = function (address, value, coreIndex) {
        
        //Set EVM ID
        if(typeof this.sendRegmapEVMID === "function")
        {
            this.sendRegmapEVMID(coreIndex);
        }

		// Get integer address
		const intAddr = (typeof address === 'string') ? parseInt(address, 16) : address;
		const intValue = (typeof value === 'string') ? parseInt(value, 16) : value;
		if (isNaN(intAddr) || isNaN(value)) {
			console.error(moduleName, `Invalid agruments provided to writeRegister(): ${address}, ${value}`)
			return;
		}

		// Build and send write register command string
		const cmd = WRITE_REGISTER_COMMAND + ' ' + intAddr.toString(16) + ' ' + intValue.toString(16);
		let jsPromise = this.sendCommand(cmd, DEFAULT_CMDTIMEOUT_MS);

		return jsPromise;
	};

	// Lookup function to get the fully qualified register name from the register address.
	CustomCodec.prototype.findRegister = function (regAddrs) {

		// Coerce argument to number
		const address = +regAddrs;
		if (isNaN(address)) {
			console.error(moduleName, `findRegister(): requires a numeric argument, recieved ${address}`);
			return;
		}

		// If no cached info, then we cannot perform the lookup
		if (!this.cachedRegisters) {
			console.error(moduleName, "findRegister(): 'cachedRegisters' is not defined. Cannot perform lookup!");
			return;
		}

		// If no register definitions (i.e. addresses) are skipped in the JSON register map file,
		// then we can simply index the array...
		let result = this.cachedRegisters[address];
		if (result && (result.addr === address)) { return result; }

		// Else...loop through register definitions looking for the matching register address...
		const matchingRegisters = this.cachedRegisters.filter(r => r.addr === address);
		const numMatches = matchingRegisters.length;
		if (numMatches !== 1) {
			console.error(moduleName, `findRegister(): found ${numMatches} registers with address = ${address}`);
		} else {
			return matchingRegisters[0];
		}
	};

	// Lookup function to get the current register value from a specified register address in the register model.
	CustomCodec.prototype.getRegisterValue = function (regAddrs) {

		// Get register value from register model...
		let regObj = this.findRegister(regAddrs);
		if (regObj) {
			let regName = this.registerModel.id + '.' + regObj.fullyQualifiedName;
			let regBind = gc.databind.registry.getBinding(regName);
			return regBind.getValue();
		}
	};

	// Lookup function to set a register value from a specified register address in the register model.
	CustomCodec.prototype.setRegisterValue = function (regAddrs, value) {

		// Get register object which contains register info
		let regObj = this.findRegister(regAddrs);

		// Update the register model with the new register data...
		if (regObj) {
			let regBind = gc.databind.registry.getBinding(this.registerModel.id + '.' + regObj.fullyQualifiedName);
			regBind.updateValue(value);
		}
	};


	////////////////////////////////// Constants //////////////////////////////////

	// NOTE: Define additional firmware commands here...
	const WRITE_REGISTER_COMMAND = 'wreg';
	const READ_REGISTER_COMMAND = 'rreg';
	const EVM_ID_COMMAND = 'id';
	const CAPTURE_CONFIG_COMMAND = "configcap";
	const WRITEPROTECTION_ENABLE_COMMAND = 'wpenprom';
	const READ_EEPROM_COMMAND = 'rprom';
	const WRITE_EEPROM_COMMAND = 'wprom';
	const CLEAR_EEPROM_COMMAND = 'cprom';
	const READ_SCB_EEPROM_COMMAND = 'rprom scb';
	const READ_PAMB_EEPROM_COMMAND = 'rprom pamb';
	const BSL_COMMAND = 'bsl';
	const STOP_CONV_COMMAND = "stop";

	const EEPROM_SIZE = 256;
	const DEFAULT_CMDTIMEOUT_MS = 2000;


	///////////////////////// CustomCodec Helper functions /////////////////////////

	// Called in connect() method to initialize codec properties to default values
	CustomCodec.prototype.initialize = function () {

		// RX and TX buffers
		this.strRxBuffer = "";
		this.strTxBuffer = "";
		this.lastMessage = {};				// Used for resolving sendCommand() promises

		// Command queueing
		this.currentState = "idle";
		this.prevState = "";
		this.pendingSendCmds = [];
		this.pendingCmdSeqNum = 0;
		this.cmdSeqNum = 0;
		this.bytesCaptured = [];

		// Info
		this.info = {};						// Placeholder for controllerInfo, evmInfo, and firmwareInfo
		this.cachedRegisters = undefined;	// Placeholder for saving register model information
		this.serialDeviceConnected = true;	// Set connected flag
		this.connectionTimeOut = undefined;
		this.logSerialData = false;

		// Get component references
		// NOTE: These proprerties must be added to the class instance at runtime!
		this.eventLog = document.querySelector('ti-widget-eventlog-view');	// Reference to eventLog object
		this.registerModel = document.querySelector('ti-model-register');	// Reference to register model object
		if (!this.registerModel) { console.error(moduleName, "No <ti-model-register> component found in this application!"); }

		// Cache register block information
		// NOTE: This function will need to be called after a 'registerDefsChanged' event.
		// We don't use an event listener here since initialize() is called multiples times during startup.
		this.cacheRegisterData();
	};

	// Cache (flattened) register data
	CustomCodec.prototype.cacheRegisterData = function () {
		try {
			const deviceIndex = 0;
			this.registerModel.getRegisterViewInformation(deviceIndex).then((deviceObj) => {
				if (!deviceObj) { throw "Register view information not found!"; }

				// Flatten register blocks
				this.cachedRegisters = [];
				deviceObj.regblocks.forEach(block => block.registers.forEach(reg => this.cachedRegisters.push(reg)));

				// Sort registers by address
				this.cachedRegisters = this.cachedRegisters.sort((registerA, registerB) => {
					if (registerA.addr > registerB.addr) { return 1; }
					else if (registerB.addr > registerA.addr) { return -1; }
					else {
						throw `Cannot have two registers with the same address (${registerA.name} and ${registerB.name}, address = ${registerA.addr})!`;
					}
				});
			});
		}
		catch (err) { console.error(moduleName, err); }
	};

	// Sends a command to the EVM firmware (with an optional timeout) and returns a promise with the response
	// CONSIDER: Can we expand this function to optionally select which decode object triggers resolve, and what value to use when resolving? ...This might require implementing our own queue
	// NOTE 1: This function returns a native JS promise. Take care when mixing it with Q promises!
	// NOTE 2: The resolved value will be whatever is contained in 'this.lastMessage'. Usually this is some JSON response from the device.
	CustomCodec.prototype.sendCommand = function (cmdString, msTimeout) {

		// Coerce to number
		msTimeout = +msTimeout || 0;

		// Debugging...
		let msg = `sendCommand('${cmdString}', ${msTimeout})`;
		gc.console.debug(moduleName, msg);

		// Add the pending command to the internal commandQueue
		// IMPORTANT: This promise is resolved when this.setState("idle") is called, and this.lastMessage will be the result
		const matchingPromiseString = "sendCommand"; 	// NOTE: Previously this was 'cmdString';
		let cmdPromise = Utilities.wrapQ(this.addCommand(matchingPromiseString, this.pendingCmdSeqNum++));
		let returnPromise = cmdPromise;

		if (msTimeout) {
			let racePromise = Utilities.promiseTimeout(msTimeout, cmdPromise)
				.then(result => {
					return result;  // cmdPromise completed
				})
				.catch(error => {
					// The cmdPromise was rejected or the timeout occurred (likely the latter)
					msg = `'${cmdString}' command was rejected with reason: '${error}'`;
					gc.console.debug(moduleName, msg);
					throw new Error(msg);
				});

			// Update the return promise pointer to racePromise (since this is the promise we want to wait for)
			returnPromise = racePromise;
		}

		// Read ID of the EVM and make sure it matches
		this.encoder(cmdString);

		return returnPromise;
	};

	// TODO: Standardize the collect commands

	// Capture data
	CustomCodec.prototype.captureConfig = function (bitRate, samplingRate) {
		samplingRate = parseInt(samplingRate);
		this.encoder(`${CAPTURE_CONFIG_COMMAND} ${bitRate.toString(16)} ${samplingRate.toString(16)}`);
	}

	CustomCodec.prototype.startCollect = function (delay, collectFlags, channelIDs, numEVMs) {

		console.log(moduleName, `starting collection from ${numEVMs} EVMs...`);

// 		let promise = this.addCommand('collect', this.pendingCmdSeqNum++);
		this.encoder(`collect ${delay} ${collectFlags} ${channelIDs} ${numEVMs}`);
		//this.dataCapture = true;

		// this.connectionTimeOut = setTimeout(() => {
		// 	this.resolveCmdPromise(/^collect/, 'Connection timed out', true);
		// 	clearTimeout(this.connectionTimeOut);
		// }, DEFAULT_CMDTIMEOUT_MS);

		// CONSIDER: Adjusting timeout based on sampleSize and data rate (in case more time is needed)?

		//this.bytesToCapture = bytes;


		// FOR TESTING WITHOUT HARDWARE -> Generate random data and resolve promise with random data array
		// setTimeout(() => {
		// 	let returnData = Array.from({ length: sampleSize }, () => Math.floor(Math.random() * 10));
		// 	self.resolveCmdPromise(/^collect/, returnData);
		// }, 1000);

// 		return promise;
	}

	CustomCodec.prototype.stopCollect = function () {
		this.encoder(STOP_CONV_COMMAND);
	}
	
	CustomCodec.prototype.updateCollectDelay = function (delayTime) {
		this.encoder(`setdelay ${delayTime}`);
	}
	
	CustomCodec.prototype.sendRegmapEVMID = function (evm) {
	    
	    if(evm === -1)
	    {
	        return Promise.resolve();
	    }
	    
	    // Get channels/addresses per EVM
        let channelIDs = document.querySelector('ti-custom-data-capture').getChannelIDs()
        
        // Get specific channel/address
        let EVMID = (channelIDs >> 4*(evm)) & 0xF;
        
	    if(EVMID_OLD != EVMID)
	    {
	        EVMID_OLD = EVMID;
    		let promise = this.addCommand('setdevice', this.pendingCmdSeqNum++);
    		this.encoder(`setdevice ${EVMID}`);
    		
    		//read all registers from new device
            const regPage = document.querySelector('ti-widget-register-page');
			if (regPage && regPage._readAllRegistersClickHdlr) { regPage._readAllRegistersClickHdlr(); }
    		
    		return promise;
	    }
    	    return Promise.resolve();
	}
	
	CustomCodec.prototype.getADCranges = async function (numEVMs, channelIDs) {
        //Get default adc range settings
        let jsonResponse = await this.sendCommand(`getadcrange ${numEVMs} ${channelIDs}`, DEFAULT_CMDTIMEOUT_MS);
        
        adcrange[0] = jsonResponse.ADCRANGEs & 0x1;
        
        if(numEVMs > 1)
        {
            adcrange[1] = (jsonResponse.ADCRANGEs >> 1) & 0x1;
        }
        if(numEVMs > 2)
        {
            adcrange[2] = (jsonResponse.ADCRANGEs >> 2) & 0x1;
        } 
        if(numEVMs > 3)
        {
            adcrange[3] = (jsonResponse.ADCRANGEs >> 3) & 0x1;
        }
        
        return 0;
	}

	// CustomCodec.prototype.resolveCapture = function (rawData) {
	// 	this.bytesCaptured = this.bytesCaptured.concat(rawData);
	// 	if (this.bytesCaptured.length >= this.bytesToCapture) {
	// 		console.log("Captured Resolved");
	// 		this.resolveCmdPromise(/^capture/, this.bytesCaptured, false);
	// 		clearTimeout(this.connectionTimeOut);
	// 		this.bytesCaptured = [];
	// 		this.dataCapture = false;
	// 		if (this.handleCaptureInterrupt) {
	// 			this.handleInterrupt();
	// 			this.handleCaptureInterrupt = false;
	// 		}
	// 	}
	// }

	// Update EVM state variable
	CustomCodec.prototype.setState = function (newState) {

		// Update state variables
		let prevState = this.currentState;
		this.currentState = newState;

		// State machine logic...
		switch (newState) {
		    
            case "collecting": 
			case "idle":

				// Handle state transistions
				if (prevState === "busy") {

					this.resolveCmdPromise(/^sendCommand/, this.lastMessage);	// Resolve sendCommand promise (if any)

				} 
				// else if (prevState === "collecting") {

				// 	// Resolve 'collect' promise
				// 	// NOTE: The scripting widget uses this resolved data for passing data to the scripting thread.
				// 	//let data = Utilities.getSafe(() => document.querySelector('ti-custom-data-capture').data);
				// // 	this.resolveCmdPromise(/^collect/, true);	// Resolve collect promise with data array (if any)

				// }
				
				// Dequeue next command (if any)...
				// NOTE: Make sure to update the internal state variable before
				// calling this function, or else the dequeuing method will fail!
				this.dequeueCmd();

				break;

			case "busy":
				break;

			// Handle invalid states...
			default:
				console.error(moduleName, `Invalid state argument in setState(): '${newState}'`);
				return;
		}
	};

	// Enqueue command
	CustomCodec.prototype.enqueueCmd = function (target, cmdString) {

		// Enque command
		this.pendingSendCmds.push({
			cmd: cmdString,
			//promise: this.addCommand(cmd, this.pendingCmdSeqNum++) // NOTE: We may not want to create a promise with each command...
		});
		gc.console.debug(moduleName, `Enquing command '${cmdString}' (${this.pendingSendCmds.length} cmds pending)`);
	}

	// Dequeue command
	CustomCodec.prototype.dequeueCmd = function () {

		// Check if we have any pending commands to dequeue
		const numPendingCmds = this.pendingSendCmds.length;
		if (numPendingCmds > 0) {
			const nextCmd = this.pendingSendCmds.shift();
			gc.console.debug(moduleName, `Dequing command '${nextCmd.cmd}' (${numPendingCmds} pending)`);
			this.encoder(nextCmd.cmd);
		}
	}

	// Resolve command promise
	CustomCodec.prototype.resolveCmdPromise = function (regex, response, errorMsg) {

		// Check if we are resolving a promise...
		let promiseResolved = false;
		let pendingCmd = this.commandQueue.first.next;
		if (!pendingCmd || !pendingCmd.command) { return false; }
		if (regex.test(pendingCmd.command)) {

			const msg = `Resolving '${pendingCmd.command}' command promise
				[seqNo: ${pendingCmd.seqNo}, error: ${errorMsg}] with ${JSON.stringify(response)}`;
			gc.console.debug(moduleName, msg);

			if (errorMsg) {
				this.setState('idle');
				this.addErrorResponse(response, pendingCmd.command, pendingCmd.seqNo);
			} else {
				this.addResponse(response, pendingCmd.command, pendingCmd.seqNo);
			}
			this.cmdSeqNum++;
			promiseResolved = true;

			// Check if command queue is empty
			if (!pendingCmd.next) { gc.console.debug(moduleName, "The commandQueue is now empty"); }
		}

		return promiseResolved;   // Notify caller if we resolved a promise
	}

	// Build firmware version number from compiler's date/time string
	CustomCodec.prototype.parseFirmwareInfo = function (idObj) {

		// Example: jsonObj
		//	{
		//		"id": {
		//			"name": "ADS7038",
		//			"version": "0.0.1.7"
		//			"date": "Mar 12 2020"
		//			"time": "13:47:00"
		//		}	
		//	}

		// Build firmware version number from compiler's date/time string
		let getTimestamp = (date, time) => {

			// Convert strings to string arrays
			let timeArray = time.split(":"); // Ex: '16','48','59'
			let dateArray = date.split(" "); // Ex: 'Feb','','7','2019'
			dateArray = dateArray.filter(entry => entry.trim() !== ''); // Remove any null entries that occur on single digit days

			// Get index of matching month abbreviaiton
			// REFERENCE: http://infocenter.arm.com/help/index.jsp?topic=/com.arm.doc.faqs/ka9555.html
			const month_names = ["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"];
			let month_index = 1 + month_names.findIndex(e => e === dateArray[0]);
			month_index = String(month_index);

			// Build timestamp string
			let year = dateArray[2];
			let month = month_index.padStart(2, '0');
			let day = dateArray[1].padStart(2, '0');
			let hour = timeArray[0].padStart(2, '0');
			let min = timeArray[1].padStart(2, '0');
			let sec = timeArray[2].padStart(2, '0');
			const timestamp = ("" + year + month + day + hour + min + sec);
			return parseInt(timestamp);
		};

		// Build info object
		let info = {};
		info.name = idObj.id.name;
		info.version = idObj.id.version;
		info.date = idObj.id.date;
		info.time = idObj.id.time;
		info.timeStamp = getTimestamp(info.date, info.time);
		return info;
	};


	/////////////////////////////// EEPROM functions ///////////////////////////////

	// Read EEPROM
	CustomCodec.prototype.readEEPROM = async function (scbEEPROM) {

		// Example:
		// {"eeprom":"host:SCB;brd_name:DC081;dev_name:scb;brd_rev:a;brd_sno:T19J14O1353;"}

		// Get EEPROM data
		const commandStr = READ_EEPROM_COMMAND + (scbEEPROM ? " scb" : "");
		let response = await this.sendCommand(commandStr, 10000);	// NOTE: This can be a lengthy operation (> 2 sec)
		//Some formware will only look for pamb, not scb, so we need to check for pamb on scb fail
		if (response.error && scbEEPROM) 
		{
		    const commandStr2 = READ_EEPROM_COMMAND + (scbEEPROM ? " pamb" : "");
		    response = await this.sendCommand(commandStr2, 10000);
		}
		//We have no EVM EEPROM, so we will build this ourselves. Creating framework here:
		if (response.error) { return; }
		else if (!response.console) {
			console.error(moduleName, "Invalid response from firmware when returning EEPROM data to GUI:", response);
			return;
		}

		// Parse EEPROM data
		let info = {};
		const cleanMsg = response.console.replace(/'/g, '').split(": ")[1]; //.match(/'(.*)'/)[1];	// Extract the data between ' ' characters 
		
		let properties = cleanMsg.split(";").filter(e => e != "");	// Convert to array, and remove blank elements
		properties.forEach(function (property) {
			if (property !== "#SOE" && property.indexOf("#EOE") === -1) {
				let x = property.split(":");						// String is formatted as "key:value"
				info[x[0]] = x[1];									// Add key-value pair to object 
			}
		});

		return info;
	};

	// Write EEPROM
	CustomCodec.prototype.writeEEPROM = async function (eepromData) {
		let cmdString, response;
		let error = false;
		try {

			// Validate input argument
			// TODO: Add additional checks for required properties.
			if (!eepromData) { throw 'No EEPROM data provied to writeEEPROM()'; }

			// Get current serial # (if one exists)
			response = await this.readEEPROM();
			const serialNumber = Utilities.getSafe(() => SCB.connectedHwInfo.evmInfo.brd_sno);

			// Helper function
			let getDateInfo = function () {
				function replaceIO(char) {
					return char === "I" ? "J" : char === "O" ? "P" : char;
				}
				const date = new Date();
				const year = date.getFullYear().toString().slice(2);
				const month = replaceIO(String.fromCharCode(65 + date.getMonth()));
				const day = String(date.getDate()).padStart(2, '0');
				const hour = replaceIO(String.fromCharCode(65 + date.getHours()));
				const minute = String(date.getMinutes()).padStart(2, '0');
				const sec = String(date.getSeconds()).padStart(2, '0');
				return year + month + day + hour + minute + sec;
			};

			// Add last modified and serial number parameters
			eepromData["last_modified"] = getDateInfo();
			eepromData["brd_sno"] = serialNumber ? serialNumber : getDateInfo();

			// Build EEPROM data string
			let dataString = "";
			for (const property in eepromData) {
				dataString += property + ":" + eepromData[property] + ";";
				if (eepromData[property].indexOf(" ") > -1) { throw "EEPROM Data cannot contain spaces"; }
			}
			if (dataString.length > EEPROM_SIZE) { throw `Data exceeds the EEPROM size of ${EEPROM_SIZE} bytes`; }

			// Disable EEPROM write protection
			cmdString = WRITEPROTECTION_ENABLE_COMMAND + " 0";
			response = await this.sendCommand(cmdString, DEFAULT_CMDTIMEOUT_MS);
			if (response.error || (response.acknowledge !== cmdString)) { throw "Could not disable EEPROM write protection"; }

			// Clear EEPROM
			cmdString = CLEAR_EEPROM_COMMAND;
			response = await this.sendCommand(cmdString, DEFAULT_CMDTIMEOUT_MS);
			if (response.error || (response.acknowledge !== cmdString)) { throw "Could not clear EEPROM"; }

			// Write to EEPROM
			cmdString = WRITE_EEPROM_COMMAND + " " + dataString;
			response = await this.sendCommand(cmdString, 10000);	// NOTE: This can be a lengthy operation (> 2 sec)
			if (response.error || (response.acknowledge !== cmdString)) { throw `Failed to write to EEPROM: ${response}`; }

			// Disable EEPROM write protection
			cmdString = WRITEPROTECTION_ENABLE_COMMAND + " 1";
			response = await this.sendCommand(cmdString, DEFAULT_CMDTIMEOUT_MS);
			if (response.error || (response.acknowledge !== cmdString)) { throw "Could not re-enable EEPROM write protection"; }

			// Readback EEPROM, update HW Info object, and test if data matches
			let readBackInfo = await this.readEEPROM();
			if (SCB && SCB.connectedHwInfo) { SCB.connectedHwInfo.evmInfo = readBackInfo; }
			if (!Utilities.shallowEqual(eepromData, readBackInfo)) { throw "Readback data does not match"; }

		} catch (err) {

			// Log error
			console.error(moduleName, err);
			error = true;
		}

		return error;
	};

	// Clear EEPROM - Used for testing only
	CustomCodec.prototype.clearEEPROM = async function () {
		let cmdString, response;
		let error = false;
		try {
			// Disable EEPROM write protection
			cmdString = WRITEPROTECTION_ENABLE_COMMAND + " 0";
			response = await this.sendCommand(cmdString, DEFAULT_CMDTIMEOUT_MS);
			if (response.error || (response.acknowledge !== cmdString)) { throw "Could not disable EEPROM write protection"; }

			// Clear EEPROM
			cmdString = CLEAR_EEPROM_COMMAND;
			response = await this.sendCommand(cmdString, DEFAULT_CMDTIMEOUT_MS);
			if (response.error || (response.acknowledge !== cmdString)) { throw "Could not clear EEPROM"; }

			// Disable EEPROM write protection
			cmdString = WRITEPROTECTION_ENABLE_COMMAND + " 1";
			response = await this.sendCommand(cmdString, DEFAULT_CMDTIMEOUT_MS);
			if (response.error || (response.acknowledge !== cmdString)) { throw "Could not re-enable EEPROM write protection"; }

			// Readback EEPROM, update HW Info object, and test if data matches
			let readBackInfo = await this.readEEPROM();
			if (readBackInfo !== undefined) { throw `clearEEPROM failed. Readback of EEPROM returned '${readBackInfo}'`; }

		} catch (err) {
			console.error(moduleName, err);
			error = true;
		}

		return error;
	}


	////////////////////////////// Scripting functions /////////////////////////////

	// TODO: Analyze and plot data



	///////////////////////////// Register CustomCodec /////////////////////////////

	// registerCustomCodec has to be called at startup in order for the codec to be recognized and used

	// You can 'chain' codecs together to make it easier to work with the data
	// coming from the device or going to the device.
	// The following codecs are provided:
	//   CR  - use this if the target data uses /n as a delimiter for framing.
	//         The CR codec will take care of buffering the incoming data so that
	//         your codec will only receive complete 'ready to use' packets
	//  If your target uses \n delimited strings, set baseCodecs = "CR"
	//  If your target uses some other format for data (e.g. binary data or custom strings) leave baseCodecs = "undefined"
	var baseCodecs;
	gc.console.debug(moduleName, "CustomCodec: about to register...");
	gc.databind.registerCustomCodec("SCB", CustomCodec, baseCodecs);

}

if (document.querySelector('#template_obj')) { initCodec(); }
else { document.addEventListener('WebComponentsReady', initCodec); }